import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.TreeMap;


/**
 * Stan Bak
 * 6/16/2009
 * 
 * VHDL-structured Maude to VHDL Translator
 * 
 * Sample of VHDL-Structured Maude:
(cycle-count)
(input 'slope 16)
(input 'slope2 16)

(output 'output 8)

(signal 'fsmState 8)
(signal 'internal 8)
(signal 'internal2 4)

(state 1 assignInt 'internal 9 2)
(state 2 assignInt 'internal2 9 3)
(state 3 conditionalGreaterSignalSignal 'slope 'slope2 4 5)
(state 4 assignInt 'output 4 0)
(state 5 assignInt 'output 5 0)
 
 */


public class Main 
{
	private static VHDL_Entity entity = null;
	private static TreeMap <String, Integer> signalToSizeMap = new TreeMap <String, Integer>(); 
	
	public static void main(String[] args)
	{
		String inputFilename = "input.txt";
		String outputFilename;
		String baseName = "SafeBehavior";
		String behavioralFilename = "beh.txt";
		
		if (args.length > 0)
			inputFilename = args[0];
		
		int periodIndex = inputFilename.indexOf('.');
		
		if (periodIndex != -1)
			baseName = inputFilename.substring(0, periodIndex);
		
		outputFilename = baseName + ".vhd";
		entity = new VHDL_Entity(baseName);
		entity.createBlank(baseName);
		
		try
		{
			BufferedReader fin = new BufferedReader(new InputStreamReader(new FileInputStream(inputFilename)));
			BufferedWriter fout = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(behavioralFilename)));
		
			initialWork(fout);
			
			for (String line = fin.readLine(); line != null; line = fin.readLine())
			{
				// trim comments
				int commentIndex = line.indexOf("---");
				
				if (commentIndex != -1)
					line = line.substring(0, commentIndex);
			
				// trim whitespace
				line = line.trim();
				
				if (line.length() == 0) // blank line
					continue;
				
				// check for parentheses
				if (line.length() < 2 || line.charAt(0) != '(' || line.charAt(line.length() - 1) != ')')
				{
					System.err.println("Error: line does not both start and end with parentheses: '" + line + "'");
					continue;
				}
				
				// trim parentheses
				int len = line.length();
				line = line.substring(1, len - 1);
				
				// finally print out the result
				processLine(line, fout);
			}
			
			finalWork(fout);
			
			fin.close();
			fout.close();
			
			// set entity behavioral file
			entity.addBehavior(behavioralFilename);
			
			// okay now print the entity
			entity.printEntity(outputFilename);
			
			System.out.println("Done!");
		}
		catch (Exception e)
		{
			System.err.println(e);
		}
		
		System.exit(0);
	}

	/**
	 * Set up entity and behavioral file
	 * @param fout the behavior file bufferedwriter to print the vhdl to
	 * @throws Exception if there's a file I/O error
	 */
	private static void initialWork(BufferedWriter fout) throws Exception 
	{
		entity.addSignal(new VHDL_Signal("clock", "std_logic", SignalDirection.IN, "'0'"));
		
		fout.write("\nprocess (clock) begin\n");
		fout.write("if rising_edge(clock) then\n");
		fout.write("case fsmState is\n");
		fout.write("\twhen 0 =>\n");
		fout.write("\t\tdone <= '0';\n\n");
		fout.write("\t\tif start = '1' then\n");
		fout.write("\t\t\tfsmState <= 1;\n");
		fout.write("\t\tend if;\n");
		
		entity.addSignal(new VHDL_Signal("start", "std_logic", SignalDirection.IN, "'0'"));
		setSignalSize("'start", 1);
		
		entity.addSignal(new VHDL_Signal("done", "std_logic", SignalDirection.OUT, "'0'"));
		setSignalSize("'done", 1);
	}
	
	/**
	 * Complete the Behavioral file
	 * @param fout the behavior file bufferedwriter to print the vhdl to
	 * @throws Exception if there's a file I/O error
	 */
	private static void finalWork(BufferedWriter fout) throws Exception 
	{
		fout.write("\twhen others =>\n");
		fout.write("end case;\n");
		fout.write("end if;\n");
		fout.write("end process;\n");
	}
	
	/**
	 * Add a transition to the output stream of the fsmState signal
	 * @param tabs the number of tabs to put before each output line
	 * @param to the destination state, should be an integer
	 * @param fout the stream to write to
	 * @throws Exception if an IO error occurs or integer parse errors
	 */
	private static void fsmTransition(int tabs, String to, BufferedWriter fout) throws Exception
	{
		int state = Integer.parseInt(to);
		
		if (state == 0) // done!
		{
			for (int x = 0; x < tabs; ++x)
				fout.write("\t");
			
			fout.write("done <= '1';\n");
		}
		
		for (int x = 0; x < tabs; ++x)
			fout.write("\t");
		
		fout.write("fsmState <= " + state + ";\n");
	}

	/**
	 * This does the actual one-to-one translations for the conversion
	 * @param line the current line, non-blank with comments and parentheses removed
	 * @param fout the behavior file bufferedwriter to print the vhdl to
	 * @throws Exception if there's a file I/O error
	 */
	private static void processLine(String line, BufferedWriter fout) throws Exception 
	{
		String tokens[] = line.split(" ");
		
		try
		{
			if (tokens.length == 1 && tokens[0].equals("cycle-count"))
				; // do nothing
			else if (tokens.length == 3 && tokens[0].equals("input") && tokens[1].length() > 1)
			{
				// (input 'slope 16)
				String name = tokens[1].substring(1);
				int len = Integer.parseInt(tokens[2]);
				setSignalSize(tokens[1], len);
				
				entity.addSignal(new VHDL_Signal(name, "signed(" + (len-1) + " downto 0)", SignalDirection.IN, "(others => '0')" ));
			}
			else if (tokens.length == 3 && tokens[0].equals("output") && tokens[1].length() > 1)
			{
				// (output 'output 8)
				String name = tokens[1].substring(1);
				int len = Integer.parseInt(tokens[2]);
				setSignalSize(tokens[1], len);
				
				entity.addSignal(new VHDL_Signal(name, "signed(" + (len-1) + " downto 0)", SignalDirection.OUT, "(others => '0')" ));
			}
			else if (tokens.length == 3 && tokens[0].equals("signal") && tokens[1].equals("'fsmState"))
			{
				// (signal 'fsmState 8)
				int len = Integer.parseInt(tokens[2]);
				setSignalSize(tokens[1], len);
				int rangeMax = (int)Math.round(Math.pow(2, len)) - 1;
				
				entity.addSignal(new VHDL_Signal("fsmState", "integer range 0 to " + rangeMax, SignalDirection.INTERNAL, "0" ));
			}
			else if (tokens.length == 3 && tokens[0].equals("signal") && tokens[1].length() > 1)
			{
				// (signal 'appple 8)
				String name = tokens[1].substring(1);
				int len = Integer.parseInt(tokens[2]);
				setSignalSize(tokens[1], len);
				
				entity.addSignal(new VHDL_Signal(name, "signed(" + (len-1) + " downto 0)", SignalDirection.INTERNAL, "(others => '0')" ));
			}
			else if (tokens.length > 4 && tokens[0].equals("state"))
			{
				int stateNum = Integer.parseInt(tokens[1]);
				fout.write("\twhen " + stateNum + " =>\n\t\t");
				
				if (tokens[2].equals("assignInt"))
				{
					String name = tokens[3].substring(1);
					fout.write(name + " <= to_signed(" + tokens[4] + ", " + getSignalSize(tokens[3]) + ");\n");
					
					fsmTransition(2,tokens[5],fout);
				}
				else if (tokens[2].equals("assignSignal")) 
				{
					String name = tokens[3].substring(1);
					String rhs = tokens[4].substring(1);
					fout.write(name + " <= to_signed(to_integer(" + rhs + "), " + getSignalSize(tokens[3]) + ");\n");
					fsmTransition(2,tokens[5],fout);
				}
				else if (tokens[2].equals("assignAddSignalInt")) // add is okay to simulate without casting 
				{
					String name = tokens[3].substring(1);
					String rhsSig = tokens[4].substring(1);
					fout.write(name + " <= " + rhsSig + " + (" + tokens[5] + ");\n");
					fsmTransition(2,tokens[6],fout);
				}
				else if (tokens[2].equals("assignAddSignalSignal")) // add is okay to simulate without casting 
				{
					String name = tokens[3].substring(1);
					String rhsSig = tokens[4].substring(1);
					String rhsSig2 = tokens[5].substring(1);
					fout.write(name + " <= " + rhsSig + " + " + rhsSig2 + ";\n");
					fsmTransition(2,tokens[6],fout);
				}
				else if (tokens[2].equals("assignMultSignalInt")) // mult needs casting for correct simulation
				{
					String name = tokens[3].substring(1);
					String rhsSig = tokens[4].substring(1);
					fout.write(name + " <= to_signed(to_integer(" + rhsSig + ") * (" + tokens[5] + "), " + getSignalSize(tokens[4]) + ");\n");
					fsmTransition(2,tokens[6],fout);
				}
				else if (tokens[2].equals("assignShiftRightSignalInt"))  // srl simulates okay without cast
				{
					String name = tokens[3].substring(1);
					String rhsSig = tokens[4].substring(1);
					fout.write(name + " <= " + rhsSig + " srl (" + tokens[5] + ");\n"); // srl = shift right logical
					fsmTransition(2,tokens[6],fout);
				}
				else if (tokens[2].equals("conditionalGreaterSignalInt")) 
				{
					String sig = tokens[3].substring(1);
					
					fout.write("if " + sig + " > " + tokens[4] + " then\n");
					fsmTransition(3,tokens[5],fout);
					fout.write("\t\telse\n");
					fsmTransition(3,tokens[6],fout);
					fout.write("\t\tend if;\n");
				}
				else if (tokens[2].equals("conditionalEqualsSignalInt")) 
				{
					String sig = tokens[3].substring(1);
					
					fout.write("if " + sig + " = " + tokens[4] + " then\n");
					fsmTransition(3,tokens[5],fout);
					fout.write("\t\telse\n");
					fsmTransition(3,tokens[6],fout);
					fout.write("\t\tend if;\n");
				}
				else if (tokens[2].equals("conditionalGreaterSignalSignal")) 
				{
					String sig = tokens[3].substring(1);
					String otherSig = tokens[4].substring(1);
					
					fout.write("if " + sig + " > " + otherSig + " then\n");
					fsmTransition(3,tokens[5],fout);
					fout.write("\t\telse\n");
					fsmTransition(3,tokens[6],fout);
					fout.write("\t\tend if;\n");
				}
				else if (tokens[2].equals("conditionalLesserSignalInt")) 
				{
					String sig = tokens[3].substring(1);
					
					fout.write("if " + sig + " < " + tokens[4] + " then\n");
					fsmTransition(3,tokens[5],fout);
					fout.write("\t\telse\n");
					fsmTransition(3,tokens[6],fout);
					fout.write("\t\tend if;\n");
				}
				else
				{
					throw new Exception("(2) Unprocessable line: '" + line + "'");
				}
			}
			else
			{
				throw new Exception("(1) Unprocessable line: '" + line + "'");
			}
		}
		catch (Exception e)
		{
			System.out.println("Parse Error on line '" + line + "'");
			throw e;
		}
	}

	/**
	 * Set the size for a signal, which can then be retrieved using getSignalValue
	 * @param signalName the name of the signal, with the Qid required '
	 * @param size the size of the associated signal
	 * @throws Exception if the signal already has a size set
	 */
	private static void setSignalSize(String signalName, int size) throws Exception 
	{
		if (signalToSizeMap.get(signalName) != null)
			throw new Exception("Signal '" + signalName + "' already has its size set.");
		
		signalToSizeMap.put(signalName, size);
	}
	
	/**
	 * Get the size of signal, previously set with setSignalSize
	 * @param signalName the name of the signal we want the size of, with the Qid-necessary '
	 * @return the size of the signal
	 * @throws Exception if the signal's size was not previously set with setSignalSize
	 */
	private static int getSignalSize(String signalName) throws Exception 
	{
		Integer i = signalToSizeMap.get(signalName);
		
		if (i == null)
			throw new Exception("Signal '" + signalName + "' does not have a set size.");
			
		return i.intValue();
	}
}



